/*
 * XYZ Library - draw.c
 *
 * Copyright (C) 2001 Woodley Packard
 * 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/keysymdef.h>
#include	<X11/Xos.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	"window_data.h"

#include	<stdio.h>
#include	<time.h>
#include	<stdarg.h>

extern Display		*display;
extern Colormap	cmap;
extern int			screen;

extern window_t	*yz_current_window;
extern window_data	*yz_d;

void	transform_point(int	*x, int	*y)
{
	*x += yz_current_window->origin_x;
	*y += yz_current_window->origin_y;
}

yzPixel2i3i(int	x, int	y, int	r, int	g, int	b)
{
	int	oldcol;
	pen_t	pt;
	
	if(!yz_current_window)return -1;
	transform_point(&x, &y);
	pt = *yz_current_window->pen;
	oldcol = yz_d->col;
	yzPenColor3i(r, g, b);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFillRectangle(display, yz_d->xw, yz_d->gc, x, y, 1, 1);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)XFillRectangle(display, yz_d->xp, yz_d->pgc, x, y, 1, 1);
	yzPenColor3i(pt.color.r, pt.color.g, pt.color.b);
	return 0;
}

yzPixel2i1c(int	x, int	y, color_t	c)
{
	return yzPixel2i3i(x, y, c.r, c.g, c.b);
}

int	do_select_pen()
{
	XGCValues	attr;
	XColor	creq;
	unsigned long	oldcolor;
	
	if(yz_current_window->pen->size>1) attr.line_width = yz_current_window->pen->size;
	else attr.line_width = 0;
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XChangeGC(display, yz_d->gc, GCLineWidth, &attr);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)XChangeGC(display, yz_d->pgc, GCLineWidth, &attr);
	
	oldcolor = yz_d->col;
	
	creq.red = yz_current_window->pen->color.r;
	creq.green = yz_current_window->pen->color.g;
	creq.blue = yz_current_window->pen->color.b;
	if(creq.red == 0 && creq.green == 0 && creq.blue == 0)
		creq.pixel = BlackPixel(display, screen);
	else if(!XAllocColor(display, cmap, &creq))
	{
		fprintf(stderr, "failed to allocate color: %d,%d,%d\n", creq.red, creq.green, creq.blue);
		creq.pixel = BlackPixel(display, screen);
	}
	yz_d->col = creq.pixel;
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
	{
		if(!XSetForeground(display, yz_d->gc, yz_d->col))
		{
			fprintf(stderr, "failed to set foreground for front buffer\n");
		}
	}
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)XSetForeground(display, yz_d->pgc, yz_d->col);
	
	if(oldcolor != BlackPixel(display, screen))
		XFreeColors(display, cmap, &oldcolor, 1, 0);
}

yzwPenColor3i(int	r, int	g, int	b)
{
	if(!yz_current_window)return -1;
	yz_current_window->pen->color.r = r;
	yz_current_window->pen->color.g = g;
	yz_current_window->pen->color.b = b;
	do_select_pen();
	return 0;
}

int	(*yzPenColor3i)(int,int,int) = yzwPenColor3i;

yzPenColor1c(color_t	c)
{
	return yzPenColor3i(c.r, c.g, c.b);
}

yzPenColor4i(int	r, int	g, int	b, int	a)
{
	return yzPenColor3i(r, g, b);
}

yzPenSize1i(int	s)
{
	if(!yz_current_window)return;
	
	yz_current_window->pen->size = s;
	do_select_pen();
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzwLine2i2i(int	x1, int	y1, int	x2, int	y2)
{
	if(!yz_current_window)return -1;
	transform_point(&x1,&y1);
	transform_point(&x2,&y2);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XDrawLine(display, yz_d->xw, yz_d->gc, x1, y1, x2, y2);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)XDrawLine(display, yz_d->xp, yz_d->pgc, x1, y1, x2, y2);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

int	(*yzLine2i2i)(int,int,int,int) = yzwLine2i2i;

yzwRect2i2i(int	x1, int	y1, int	x2, int	y2)
{
	if(!yz_current_window)return -1;
	transform_point(&x1,&y1);
	transform_point(&x2,&y2);
	if(x1>x2 || y1>y2)return;
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
	{
		//fprintf(stderr, "trying to fill in FRONT buffer; %d,%d, size %d,%d\n",
					//x1, y1, x2-x1+1, y2-y1 + 1);
		if(!XFillRectangle(display, yz_d->xw, yz_d->gc, x1, y1, x2-x1+1, y2-y1+1))
			fprintf(stderr, "error in xfillrectangle\n");
	}
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
	{
		//fprintf(stderr, "trying to fill in BACK buffer; %d,%d, size %d,%d\n",
					//x1, y1, x2-x1+1, y2-y1 + 1);
		XFillRectangle(display, yz_d->xp, yz_d->pgc, x1, y1, x2-x1+1, y2-y1+1);
	}
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

int	(*yzRect2i2i)(int,int,int,int) = yzwRect2i2i;

yzOutlineRect2i2i(int	x1, int	y1, int	x2, int	y2)
{
	yzLine2i2i(x1, y1, x2, y1);
	yzLine2i2i(x1, y1, x1, y2);
	yzLine2i2i(x2, y2, x2, y1);
	yzLine2i2i(x2, y2, x1, y2);
	return 0;
}

yzClear()
{
	if(!yz_current_window)return -1;
	yzRect(0, 0, yz_current_window->width-1, yz_current_window->height-1);
	return 0;
}

yzFillCircle3i(int	x, int	y, int	r)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XFillArc(display, yz_d->xw, yz_d->gc, x-r, y-r, r*2, r*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XFillArc(display, yz_d->xp, yz_d->pgc, x-r, y-r, r*2, r*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzwCircle3i(int	x, int	y, int	r)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XDrawArc(display, yz_d->xw, yz_d->gc, x-r, y-r, r*2, r*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XDrawArc(display, yz_d->xp, yz_d->pgc, x-r, y-r, r*2, r*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

int	(*yzCircle3i)(int x, int y, int r) = yzwCircle3i;

yzFillOval4i(int	x, int	y, int	rx, int	ry)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XFillArc(display, yz_d->xw, yz_d->gc, x-rx, y-ry, rx*2, ry*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XFillArc(display, yz_d->xp, yz_d->pgc, x-rx, y-ry, rx*2, ry*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzOval4i(int	x, int	y, int	rx, int	ry)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XDrawArc(display, yz_d->xw, yz_d->gc, x-rx, y-ry, rx*2, ry*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XDrawArc(display, yz_d->xp, yz_d->pgc, x-rx, y-ry, rx*2, ry*2, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzFillOvalInRect2i2i(int	x, int	y, int	x2, int	y2)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	transform_point(&x2,&y2);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XFillArc(display, yz_d->xw, yz_d->gc, x, y, x2-x, y2-y, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XFillArc(display, yz_d->xp, yz_d->pgc, x, y, x2-x, y2-y, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzOvalInRect2i2i(int	x, int	y, int	x2, int	y2)
{
	if(!yz_current_window)return -1;
	transform_point(&x,&y);
	transform_point(&x2,&y2);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XDrawArc(display, yz_d->xw, yz_d->gc, x, y, x2-x, y2-y, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XDrawArc(display, yz_d->xp, yz_d->pgc, x, y, x2-x, y2-y, 0, 64*360);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
}

yzPolygon(int	n, ...)
{
	va_list	va;
	XPoint	*p;
	int		i, x, y;

	p = (XPoint*)calloc(sizeof(XPoint), n);
	va_start(va, n);
	for(i=0;i<n;i++)
	{
		x = va_arg(va, int);
		y = va_arg(va, int);
		transform_point(&x,&y);
		p[i].x = x; p[i].y = y;
	}
	va_end(va);
	if(!yz_current_window)
	{
		free(p);
		return -1;
	}
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XFillPolygon(display, yz_d->xw, yz_d->gc, p, n, Complex, CoordModeOrigin);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XFillPolygon(display, yz_d->xp, yz_d->pgc, p, n, Complex, CoordModeOrigin);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	free(p);
	return 0;
}

yzPolygon1i2p(int	n, int	*x, int	*y)
{
	XPoint	*p;
	int		i, X, Y;

	p = (XPoint*)malloc(sizeof(XPoint) * n);
	for(i=0;i<n;i++)
	{
		X = x[i]; Y = y[i];
		transform_point(&X,&Y);
		p[i].x = X;
		p[i].y = Y;
	}
	if(!yz_current_window)
	{
		free(p);
		return -1;
	}
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XFillPolygon(display, yz_d->xw, yz_d->gc, p, n, Complex, CoordModeOrigin);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XFillPolygon(display, yz_d->xp, yz_d->pgc, p, n, Complex, CoordModeOrigin);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	free(p);
	return 0;
}

yzwText2i1p1i(int	x, int	y, char	*str, int	n)
{
	if(!yz_current_window)return -1;
	if(n==0 || str==0)return 0;
	transform_point(&x, &y);

#ifdef USE_PANGO
	if (yz_d->buffer_mode & YZ_BUFFER_FRONT) {
		pango_render_text(yz_d->xw, str, n, x, y, yz_d->gc);
	}
	if (yz_d->buffer_mode & YZ_BUFFER_BACK) {
		pango_render_text(yz_d->xp, str, n, x, y, yz_d->pgc);
    }

#else
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		Xutf8DrawString(display, yz_d->xw, yz_d->fnt, yz_d->gc, x, y, str, n);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		Xutf8DrawString(display, yz_d->xp, yz_d->fnt, yz_d->pgc, x, y, str, n);
#endif
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;

/*
	int		len, i;
	wchar_t	*wide;
	XChar2b	*c2b;

	if(!yz_current_window)return -1;
	if(n==0 || str==0)return 0;
	transform_point(&x, &y);
	wide = malloc(sizeof(wchar_t)*n);
	len = mbsnrtowcs(wide, &str, n, n, 0);
	if(len<=0)
	{
		fprintf(stderr, "error: unable to convert `%.*s' to a wide-char string\n", n, str);
		perror("maybe forgot to setlocale(LC_ALL, \"\")? mbsnrtowcs");
		free(wide);
		if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
			XDrawString(display, yz_d->xw, yz_d->gc, x, y, str, n);
		if(yz_d->buffer_mode & YZ_BUFFER_BACK)
			XDrawString(display, yz_d->xp, yz_d->pgc, x, y, str, n);
	}
	c2b = malloc(sizeof(XChar2b)*n);
	for(i=0;i<len;i++)
	{
		//fprintf(stderr, "wide[%d] = %8.8X\n", i, wide[i]);
		c2b[i].byte2 = wide[i] & 0x00FF;
		c2b[i].byte1 = ((wide[i] & 0xFF00) >> 8);
	}
	free(wide);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		XDrawString16(display, yz_d->xw, yz_d->gc, x, y, c2b, len);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
		XDrawString16(display, yz_d->xp, yz_d->pgc, x, y, c2b, len);
	free(c2b);
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XFlush(display);
	return 0;
*/
}

int	(*yzText2i1p1i)(int x, int y, char	*ptr, int	len) = yzwText2i1p1i;

yzText2i1s(int	x, int	y, char	*str)
{
	return yzText2i1p1i(x, y, str, strlen(str));
}

#ifndef USE_PANGO
int	font_string_width(XFontSet	fnt, char	*s, int	n)
{
	XRectangle	ink, logic;

	Xutf8TextExtents(fnt, s, n, &ink, &logic);
	return logic.width;

/*
	int		len, width, i;
	XChar2b	*c2b;
	wchar_t	*wide;

	if(n==0 || s==0)return 0;
	wide = malloc(sizeof(wchar_t)*n);
	len = mbsnrtowcs(wide, &s, n, n, 0);
	if(len<=0)
	{
		fprintf(stderr, "error: unable to convert `%.*s' to a wide-char string\n", n, s);
		perror("maybe forgot to setlocale(LC_ALL, \"\")? mbsnrtowcs");
		free(wide);
		return XTextWidth(fnt, s, n);
	}
	c2b = malloc(sizeof(XChar2b)*n);
	for(i=0;i<len;i++)
	{
		c2b[i].byte2 = wide[i] & 0x00FF;
		c2b[i].byte1 = ((wide[i] & 0xFF00) >> 8);
	}
	free(wide);
	width = XTextWidth16(fnt, c2b, len);
	free(c2b);
	return width;
*/
}
#endif

int	non_window_default_line_height()
{
	if(!display)init_xyz();
	extern font_t	*no_window_font;
	if(no_window_font)
	{
		struct font_data	*fd = no_window_font->impl_data;
		return default_line_height(fd->fd);
	}
	else return -1;
}

int	yzwDefaultLineHeight()
{
	if(!yz_current_window)
		return non_window_default_line_height();
	return default_line_height(yz_d->fnt);
}

int	yzwStringSize(char	*s, int	n)
{
	if(!yz_current_window)return non_window_string_length(s, n);

	return font_string_width(yz_d->fnt, s, n);
}

int	(*yzStringSize)(char	*ptr, int	length) = yzwStringSize;
int (*yzDefaultLineHeight)() = yzwDefaultLineHeight;

int	yzLetterSize(int	l)
{
	char	c;
	
	if(!yz_current_window)return 0;
	c = l;
	return font_string_width(yz_d->fnt, &c, 1);
}

int	yzGetDepth()
{
	int		bits;
	
	if(!yz_current_window)return -1;
	bits = DefaultDepth(display, screen);
	if(bits==15)bits=16;
	if(bits==24)bits=32;
	return bits;
}

int	yzDisplayBits5i1p(int	x, int	y, int	w, int	h, int	rb, void	*d)
{
	XImage	im;
	int		bits;
	
	if(!yz_current_window)return -1;
	transform_point(&x, &y);
	bits = DefaultDepth(display, screen);
	if(bits==24)bits=32;
	im.width = w; im.height = h; im.xoffset = 0;
	im.format = ZPixmap; im.data = d;
	im.byte_order = MSBFirst;
	im.bitmap_unit = bits;
	im.bitmap_bit_order = MSBFirst;
	im.bitmap_pad = bits;
	im.depth = DefaultDepth(display, screen);
	im.bytes_per_line = rb;
	im.bits_per_pixel = bits;
	im.red_mask  = 0x00FF0000;
	im.green_mask= 0x0000FF00;
	im.blue_mask = 0x000000FF;
	XInitImage(&im);
	
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
	{
		XPutImage(display, yz_d->xp, yz_d->pgc, &im, 0, 0, x, y, w, h);
		if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
			XCopyArea(display, yz_d->xp, yz_d->xw, yz_d->gc, x, y, w, h, x, y);
	}
	else
	{
		XPutImage(display, yz_d->xw, yz_d->gc, &im, 0, 0, x, y, w, h);
	}
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
	{
		XFlush(display);
		XSync(display, 0);
	}
	return 0;
}

int	yzCopyRect2i2i2i(int	srcx, int	srcy, int	dstx, int	dsty, int	width, int	height)
{
	int	r;
	
	if(!yz_current_window)return -1;
	
	transform_point(&srcx, &srcy);
	transform_point(&dstx, &dsty);
	if(yz_d->buffer_mode & YZ_BUFFER_BACK)
	{
		// copy stuff in back buffer, then copy to front
		r = XCopyArea(display, yz_d->xp, yz_d->xp, yz_d->pgc, srcx, srcy, width, height, dstx, dsty);
		if(r<0)perror("XCopyArea from pix to pix failed");
		if(r<0)fprintf(stderr, "result was %d\n", r);
		if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
		{
			r = XCopyArea(display, yz_d->xp, yz_d->xw, yz_d->gc, dstx, dsty, width, height, dstx, dsty);
			if(r<0)perror("XCopyArea from pix to win failed");
			if(r<0)fprintf(stderr, "result was %d\n", r);
		}
	}
	else
	{
		r = XCopyArea(display, yz_d->xw, yz_d->xw, yz_d->gc, srcx, srcy, width, height, dstx, dsty);
		if(r<0)perror("XCopyArea from win to win failed");
		if(r<0)fprintf(stderr, "result was %d\n", r);
	}
	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)
	{
		XFlush(display);
		XSync(display, 0);
	}
	return 0;
}

void	*yzGetBackBuffer(int	*pitch)
{
	if(!yz_current_window)return 0;
	
	//fprintf(stderr, "yzGetBackBuffer: width = %d, DefaultDepth = %d\n", yz_current_window->width,
		//DefaultDepth(display, screen));
	if(pitch)*pitch = yz_current_window->width * ((DefaultDepth(display, screen)+7)/8);
	return yz_d->buffer;
}
