/*
 * XYZ Library - events.c
 *
 * Copyright (C) 2001 Woodley Packard
 * 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/keysymdef.h>
#include	<X11/Xos.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	"window_data.h"

#include	<stdio.h>
#include	<time.h>

extern Display		*display;
extern Colormap	cmap;
extern int			screen;

extern window_t	*yz_current_window;
extern window_data	*yz_d;

extern int			__xyz_window_list_n;
extern window_t	**__xyz_window_list;

void	process_events();

static window_t	*xyz_lookup_x_window(Window	w)
{
	int	i;

	for(i=0;i<__xyz_window_list_n;i++)
		if( ((window_data*)__xyz_window_list[i]->impl_data)->xw == w )
			return __xyz_window_list[i];
	return 0;
}

event_t	dequeue_window_event(window_t	*w);

int	yzFlushAllEvents()
{
	int	i;

	for(i=0;i<__xyz_window_list_n;i++)
		flush_window_events(__xyz_window_list[i]);
}

flush_window_events(window_t	*w)
{
	event_t	ev;

	do { ev = dequeue_window_event(w); } while(ev.type != YZ_IDLE);
}

int	yzQueryPointer(int	*x, int	*y)
{
	int	ox, oy, ob, res = 0;
	Window	rw, cw;
	int		rx, ry, wx, wy, mask, bm;

	// this should use XQueryPointer().
	if(!yz_current_window)return -1;

	/*if(!XQueryPointer(display, yz_d->xw, &rw, &cw, &rx, &ry, &wx, &wy, &mask))return -1;
	if(wx <0 || wy<0 || wx>=yz_current_window->width || wy>=yz_current_window->height)
	{
		fprintf(stderr, "wx = %d, wy = %d\n", wx, wy);
		if(x)*x = wx;
		if(y)*y = wy;
		bm = mask & (Mod1Mask | Mod2Mask | Mod3Mask);
		if(yz_d->x != wx || yz_d->y != wy)
			yz_d->x = wx, yz_d->y = wy, res = 1;
		if(bm != yz_d->b)
		{
			yz_d->b = bm;
			if(bm)return 2;
			else return 3;
		}
		else return res;
	}
	else return -1;*/

	ox = yz_d->x; oy = yz_d->y; ob = yz_d->b;
	process_events();
	if(yz_d->x<0 || yz_d->x>yz_current_window->width ||
	   yz_d->y<0 || yz_d->y>yz_current_window->height)
	{
		if(!yz_d->b && !ob)return -1;
	}
	if(ox != yz_d->x || oy != yz_d->y)
		res = 1;
	if(yz_d->b != ob)
		res = 2 + ob;

	if(x)*x = yz_d->x;
	if(y)*y = yz_d->y;

	return res;
}

queue_window_event(window_t	*w, event_t	ev)
{
	window_data	*d;
	event_queue	*e, *s;
	
	if(!w)return -1;
	d = w->impl_data;
	if(ev.type==YZ_MOUSE_DRAG && d->evq && d->evq->ev.type==YZ_MOUSE_DRAG)
	{
		// combine 'drag' events that haven't been received yet
		d->evq->ev.x2 = ev.x2;
		d->evq->ev.y2 = ev.y2;
		return 0;
	}
	if(d->evqd>32)dequeue_window_event(w);
	e = (event_queue*)malloc(sizeof(event_queue));
	e->ev = ev; e->next = 0;
	if(d->evq)
	{
		s = d->evq;
		while(s->next)s = s->next;
		e->prev = s;
		s->next = e;
	}
	else
	{
		e->prev = 0;
		d->evq = e;
	}
	d->evqd++;
	return 0;
}

event_t	dequeue_window_event(window_t	*w)
{
	event_t	ev;
	event_queue	*s;
	window_data	*d;
	
	ev.type = YZ_IDLE;
	if(!w)return ev;
	d = w->impl_data;
	if(!d->evq)return ev;
	s = d->evq;
	if(s->next)
		s->next->prev = 0;
	d->evq = s->next;
	ev = s->ev;
	free((char*)s);
	d->evqd--;
	return ev;
}

xlatekeycode_nomods(int	code)
{
	char	*keystring;
	KeySym	sym;
	
	sym = XKeycodeToKeysym(display, code, 0);
	keystring = XKeysymToString(sym);
	if(!keystring)return YZ_KEYCODE_UNKNOWN;
	if(!keystring[1])return keystring[0];
	if(!strcmp(keystring, "space"))return ' ';
	if(!strcmp(keystring, "Tab"))return '	';
	if(!strcmp(keystring, "grave"))return '`';
	if(!strcmp(keystring, "minus"))return '-';
	if(!strcmp(keystring, "equal"))return '=';
	if(!strcmp(keystring, "slash"))return '/';
	if(!strcmp(keystring, "bracketleft"))return '[';
	if(!strcmp(keystring, "bracketright"))return ']';
	if(!strcmp(keystring, "semicolon"))return ';';
	if(!strcmp(keystring, "apostrophe"))return '\'';
	if(!strcmp(keystring, "period"))return '.';
	if(!strcmp(keystring, "comma"))return ',';
	if(!strcmp(keystring, "backslash"))return '\\';
	if(!strcmp(keystring, "BackSpace"))return 8;
	if(!strcmp(keystring, "Return"))return '\n';
	if(!strcmp(keystring, "KP_Enter"))return 3;
	if(!strcmp(keystring, "Up"))return YZ_KEYCODE_UP_ARROW;
	if(!strcmp(keystring, "Left"))return YZ_KEYCODE_LEFT_ARROW;
	if(!strcmp(keystring, "Right"))return YZ_KEYCODE_RIGHT_ARROW;
	if(!strcmp(keystring, "Down"))return YZ_KEYCODE_DOWN_ARROW;
	if(!strncmp(keystring, "Control", 7))return YZ_KEYCODE_CONTROL;
	if(!strncmp(keystring, "Shift", 5))return YZ_KEYCODE_SHIFT;
	if(!strncmp(keystring, "Alt", 3))return YZ_KEYCODE_ALT;
	fprintf(stderr, "unknown XYZ keycode: %s\n", keystring);
	return YZ_KEYCODE_UNKNOWN;
}

xlatekeycode(int	code)
{
	int	keynomod = xlatekeycode_nomods(code);
	
	if(yzQueryKey(YZ_KEYCODE_SHIFT))
	{
		if(keynomod>='a' && keynomod<='z')
			keynomod+='A'-'a';
		if(keynomod=='-')keynomod='_';
		if(keynomod=='=')keynomod='+';
		if(keynomod=='1')keynomod='!';
		if(keynomod=='2')keynomod='@';
		if(keynomod=='3')keynomod='#';
		if(keynomod=='4')keynomod='$';
		if(keynomod=='5')keynomod='%';
		if(keynomod=='6')keynomod='^';
		if(keynomod=='7')keynomod='&';
		if(keynomod=='8')keynomod='*';
		if(keynomod=='9')keynomod='(';
		if(keynomod=='0')keynomod=')';
		if(keynomod=='/')keynomod='?';
		if(keynomod=='\\')keynomod='|';
		if(keynomod=='\'')keynomod='\"';
		if(keynomod=='[')keynomod='{';
		if(keynomod==']')keynomod='}';
		if(keynomod==';')keynomod=':';
		if(keynomod==',')keynomod='<';
		if(keynomod=='.')keynomod='>';
		if(keynomod=='`')keynomod='~';
	}
	return keynomod;
}

// given point x,y in yz_current_window coords,
// determine what window x,y is in and calculate local coords
window_t	*yzLocateWindow(int	x, int	y, int	*localx, int	*localy)
{
	Window	src, win, child;
	int		X, Y;

	if(!yz_current_window)return 0;

	win = yz_d->xw;
	child = RootWindow(display, screen);

	do
	{
		src = win;
		win = child;

		if(!XTranslateCoordinates(display, src, win, x, y, &X, &Y, &child))
			return 0;
		x = X; y = Y;
	} while(child != None);

	if(localx)*localx = x;
	if(localy)*localy = y;

	return xyz_lookup_x_window(win);
}

void	process_events()
{
	XEvent	xev;
	event_t	ev, resize_ev;
	Window	xw;
	window_t	*w, *sw, *resizew = 0;
	window_data	*d, *sd;

	resize_ev.type=YZ_IDLE;
	while(XCheckTypedEvent(display, ClientMessage, &xev))
	{
		//printf("client message\n");
		xw = xev.xany.window;
		w = xyz_lookup_x_window(xw);
		if(!w)continue;
		ev.type = YZ_WINDOW_CLOSE;
		queue_window_event(w, ev);
	}

	while(XCheckMaskEvent(display, XYZ_EVM, &xev))
	{
		xw = xev.xany.window;
		w = xyz_lookup_x_window(xw);
		if(!w)
		{
			//fprintf(stderr, "lookup for event found no window.\n");
			continue;
		}
		//else fprintf(stderr, "lookup for event found window %s\n", w);
		d = w->impl_data;
		ev.button = 0;
		switch(xev.type)
		{
			case	MotionNotify:
				if(d->b)	// if the button's down, generate a drag
				{
					ev.x = d->x; ev.y = d->y;
					ev.x2 = xev.xmotion.x;
					ev.y2 = xev.xmotion.y;
					ev.button = d->b;
					ev.type = YZ_MOUSE_DRAG;
					queue_window_event(w, ev);
				}
				d->x = xev.xmotion.x;
				d->y = xev.xmotion.y;
				break;
			case	ButtonPress:
				d->b = xev.xbutton.button;
				ev.x = xev.xbutton.x;
				ev.y = xev.xbutton.y;
				ev.button = xev.xbutton.button;
				ev.type = YZ_MOUSE_DOWN;
				queue_window_event(w, ev);
				return;	// we only want one of these per process
			case	ButtonRelease:
				d->b = 0;
				ev.x = xev.xbutton.x;
				ev.y = xev.xbutton.y;
				ev.button = xev.xbutton.button;
				ev.type = YZ_MOUSE_UP;
				queue_window_event(w, ev);
				return;	// we only want one of these per process
			case	KeyPress:
				ev.key = xlatekeycode(xev.xkey.keycode);
				ev.type = YZ_KEY_DOWN;
				queue_window_event(w, ev);
				break;
			case	KeyRelease:
				ev.key = xlatekeycode(xev.xkey.keycode);
				ev.type = YZ_KEY_UP;
				queue_window_event(w, ev);
				break;
			case	LeaveNotify:
				d->x = -1; d->y = -1;
				ev.type = YZ_WINDOW_DEACTIVATE;
				queue_window_event(w, ev);
				break;
			case	EnterNotify:
				ev.type = YZ_WINDOW_ACTIVATE;
				queue_window_event(w, ev);
				break;
			case	Expose:
				sw = yz_current_window;
				sd = yz_d;
				yz_current_window = w;
				yz_d = w->impl_data;
				yzPushClipRect(0,0,1,1);
				do_set_clip_rect(0, 0, w->width-1, w->height-1);
				XCopyArea(display, d->xp, d->xw, d->gc,
							0, 0, w->width,
							w->height, 0, 0);
				yzPopClipRect();
				yz_current_window = sw;
				yz_d = sd;
				break;
			case	ConfigureNotify:
				if(xev.xconfigure.width != w->width ||
					xev.xconfigure.height != w->height)
				{
					ev.type = YZ_WINDOW_RESIZE;
					ev.x = xev.xconfigure.width;
					ev.y = xev.xconfigure.height;
					if(resizew && resizew != w)
						queue_window_event(resizew, resize_ev);
					resize_ev = ev;
					resizew = w;
				}
				// else twas a 'move' event
				break;
			//default:	fprintf(stderr, "unknown Xlib event type %d\n", xev.type);	break;
		}
	}
	if(resize_ev.type==YZ_WINDOW_RESIZE)
		queue_window_event(resizew, resize_ev);
}

int	yzAreEventsWaiting()
{
	if(!yz_current_window)return 0;
	process_events();
	if(yz_d->evq)return 1;
	return 0;
}

resize_back_pixmap(window_t	*w, int	newx, int	newy)
{
	Pixmap		newpm;
	GC			ngc;
	XGCValues	val;
	window_data	*d = w->impl_data;
	int			r;

	newpm = make_x_pixmap(w->impl_data, newx, newy);
	val.function = GXcopy;
	ngc = XCreateGC(display, newpm, GCFunction, &val);
	//if(d->fnt)XSetFont(display, ngc, d->fnt->fid);
	r = XCopyArea(display, d->xp, newpm, ngc, 0, 0, w->width, w->height, 0, 0);
	if(r<0)fprintf(stderr, "XCopyArea for resize back pixmap failed\n");
	XFreePixmap(display, d->xp);
	XFreeGC(display, d->pgc);
	d->xp = newpm;
	d->pgc = ngc;
}

static struct timeval future_time(int	msec)
{
	struct timeval endtime;
	int		some;

	gettimeofday(&endtime, 0);
	while(msec)
	{
		if(msec>1000) { some = 1000; msec -= 1000; }
		else { some = msec; msec = 0; }
		endtime.tv_usec += 1000*some;
		if(endtime.tv_usec > 1000000)
		{
			endtime.tv_usec -= 1000000; endtime.tv_sec++;
		}
	}

	return endtime;
}

static int	until_time(struct timeval	endtime)
{
	int	msec;
	struct timeval	nowtime;

	gettimeofday(&nowtime, 0);
	msec = 1000 * (endtime.tv_sec - nowtime.tv_sec);
	msec += (endtime.tv_usec - nowtime.tv_usec) / 1000;

	return msec;
}

event_t	yzGetUnfilteredEvent(int	msec)
{
	event_t	ev;
	struct timeval	endtime, nowtime;
	
	ev.type = YZ_IDLE;
	if(!yz_current_window || msec<0)return ev;
	endtime = future_time(msec);
	do
	{
		process_events();
		ev = dequeue_window_event(yz_current_window);
		if(ev.type==YZ_WINDOW_RESIZE)
		{
			resize_back_pixmap(yz_current_window, ev.x, ev.y);
			yz_current_window->width = ev.x;
			yz_current_window->height = ev.y;
			extend_clip_on_resize();
		}
		msec = until_time(endtime);
		if(ev.type != YZ_IDLE)break;
		if(msec > 100)usleep(100000);
		else if(msec>10)usleep(10000);
		gettimeofday(&nowtime, 0);
	} while(!(nowtime.tv_sec >= endtime.tv_sec || (nowtime.tv_sec==endtime.tv_sec && nowtime.tv_usec >= endtime.tv_usec)));
	return ev;
}

event_t	yzGetEvent(int	msec)
{
	event_t	ev;
	event_filter	*f;
	struct timeval	endtime;
	
	ev.type = YZ_IDLE;
	if(!yz_current_window)return ev;
	endtime = future_time(msec);
	do
	{
		msec = until_time(endtime);
		ev = yzGetUnfilteredEvent(msec);
		if(ev.type != YZ_IDLE)
		{
			f = yz_d->evf;
			while(f)
			{
				if(f->filter(&ev))break;	// it got snarfed
				f = f->next;
			}
			if(!f)return ev;	// nobody snarfed it - so we return it.
		}
	} while(msec>0);
	ev.type = YZ_IDLE;
	return ev;
}

event_t	yzGetGlobalEvent(int	msec, window_t	**wptr)
{
	event_t	ev;
	event_filter	*f;
	struct timeval	endtime, nowtime;
	window_t	*w;
	int			i;

	ev.type = YZ_IDLE;
	endtime = future_time(msec);
	do
	{
		for(i=0;i<__xyz_window_list_n;i++)
		{
			yzSelectWindow(__xyz_window_list[i]);
			ev = yzGetEvent(0);
			if(ev.type != YZ_IDLE)
			{
				*wptr = __xyz_window_list[i];
				goto got;
			}
		}
		msec = until_time(endtime);
		if(msec>100)usleep(100000);
		if(msec>10)usleep (10000);
		ev.type = YZ_IDLE;
	} while(msec>0);

	if(display && !__xyz_window_list_n)
		XSync(display, True);

	got:
	return ev;
}

event_t	yzGetGlobalEventAndSelect(window_t	**wptr, int	nfds, fd_set	rfds, int msec)
{
	if(!display)init_xyz();
	int xfd = ConnectionNumber(display);
	FD_SET(xfd, &rfds);
	if(xfd+1 > nfds)nfds = xfd+1;
	struct timeval	tv;
	tv.tv_sec = msec / 1000;
	tv.tv_usec = (msec%1000)*1000;
	// wait until we're pretty sure something interesting has happened
	if(!XPending(display))
	{
		//printf("selecting\n");
		int r = select(nfds, &rfds, NULL, NULL, &tv);
		//printf(" = %d\n", r);
		//for(r=0;r<nfds;r++)
			//printf("fd %d -> %d\n", r, FD_ISSET(r, &rfds));
	}
	// yield, possibly with an interesting event
	return yzGetGlobalEvent(0, wptr);
}

int	yzQueryKey(int	key)
{
	char	keym[32], *k;
	char	str[15];
	KeySym	sym;
	int		code;
	
	if(!yz_current_window)return 0;
	XQueryKeymap(display, keym);
	if(key<128)
	{
		if(key==' ')strcpy(str, "space");
		else if(key=='/')strcpy(str, "slash");
		else if(key=='\n')strcpy(str, "Return");
		else if(key==3)strcpy(str, "KP_Enter");
		else { str[0]=key; str[1]=0; }
	}
	else switch(key)
	{
		case	YZ_KEYCODE_LEFT_ARROW:	strcpy(str, "Left"); break;
		case	YZ_KEYCODE_RIGHT_ARROW:	strcpy(str, "Right"); break;
		case	YZ_KEYCODE_UP_ARROW:	strcpy(str, "Up"); break;
		case	YZ_KEYCODE_DOWN_ARROW:	strcpy(str, "Down"); break;
		case	YZ_KEYCODE_CONTROL:	strcpy(str, "Control_L"); break;
		case	YZ_KEYCODE_SHIFT:	strcpy(str, "Shift_L"); break;
		case	YZ_KEYCODE_ALT:	strcpy(str, "Alt_L"); break;
		default:		strcpy(str, "moogoo"); break;
	}
	sym = XStringToKeysym(str);
	code = XKeysymToKeycode(display, sym);
	k = keym;
	while(code >= 8)
	{
		code -= 8;
		k++;
	}
	while(code>0)
	{
		code--;
		(*k) = (*k)>>1;
	}
	return (*k) & 1;
}

yzAddEventFilter(int	(*filter)(event_t*))
{
	event_filter	*f, *s;
	
	if(!yz_current_window)return -1;
	f = (event_filter*)malloc(sizeof(event_filter));
	f->filter = filter;
	f->next = 0;
	if(yz_d->evf)
	{
		s = yz_d->evf;
		while(s->next)s = s->next;
		f->prev = s;
		s->next = f;
	}
	else
	{
		f->prev = 0;
		yz_d->evf = f;
	}
	return 0;
}

yzRemoveEventFilter(int	(*filter)(event_t*))
{
	event_filter	*f;
	
	if(!yz_current_window)return -1;
	f = yz_d->evf;
	if(!f)return -1;
	if(f->filter == filter)
	{
		if(f->next)f->next->prev = 0;
		yz_d->evf = f->next;
		free((char*)f);
		return 0;
	}
	while(f->next)
	{
		f = f->next;
		if(f->filter == filter)
		{
			if(f->next)f->next->prev = f->prev;
			if(f->prev)f->prev->next = f->next;
			free((char*)f);
			return 0;
		}
	}
	return -1;
}
