/*
 * XYZ Library - draw.c
 *
 * Copyright (C) 2001 Woodley Packard
 * 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef USE_PANGO

#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/keysymdef.h>
#include	<X11/Xos.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	"window_data.h"

#include	<stdio.h>
#include	<time.h>
#include	<stdarg.h>
#include	<stdlib.h>

extern Display		*display;
extern Colormap	cmap;
extern int			screen;

extern window_t	*yz_current_window;
extern window_data	*yz_d;

font_t		*no_window_font = 0;
XFontSet	no_window_x11_font = 0;

font_t	*alloc_font()
{
	font_t		*f;
	font_data	*fd;
	
	f = malloc(sizeof(font_t));
	fd = f->impl_data = malloc(sizeof(struct font_data));
	fd->auto_dealloc = 1;
	fd->refcnt=1;

	return f;
}

free_font(font_t	*fnt)
{
	font_data	*fi;

	fi = (font_data*)fnt->impl_data;
	fi->refcnt--;
	if(fi->refcnt>0)return 0;

	XFreeFontSet(display, fi->xfs);

	free(fnt->impl_data);
	free(fnt);
}

XFontSet	x11_fs_for_font(font_t	*f)
{
	XFontSet	fs = 0;
	int			count;
	char		*xname, xfd[1024], *weight="medium", *slant="r", **missing;

	if(!display)init_xyz();

	if(f->attr == 'b')weight = "bold";
	if(f->attr == 'i')slant = "i";
	if(f->attr == 'z'){weight="bold";slant = "i";}
	xname = ((struct font_data*)f->impl_data)->x_fname;

#define TRY_XFD(display, xfd) count=0; fs=XCreateFontSet(display, xfd, &missing, &count, 0); if(!count)return fs;
#define	TRY(reit, pattern, ...)	\
	sprintf(xfd, pattern "-iso10646-1", __VA_ARGS__); fs=TRY_XFD(display, xfd); if(fs)return fs;\
	sprintf(xfd, pattern,  __VA_ARGS__); fs=TRY_XFD(display, xfd);if(fs)return fs;\
	if(slant=="i" && reit) { slant="o"; \
		sprintf(xfd, pattern "-iso10646-1",__VA_ARGS__);fs=TRY_XFD(display,xfd);if(fs)return fs;\
		sprintf(xfd, pattern,  __VA_ARGS__); fs = TRY_XFD(display, xfd);	if(fs)return fs;\
		slant="i"; }

	TRY(1, "*-%s-%s-%s-normal-*-%d-*", xname, weight, slant, f->size);
	TRY(1, "*-%s-medium-r-normal-*-%d-*", xname, f->size);
	TRY(1, "*-%s-%s-%s-normal-*", xname, weight, slant);
	TRY(1, "*-%s-medium-r-normal-*", xname);
	TRY(1, "*-%s-*", xname);

#undef TRY

	return 0;
}

font_t	*font_from_name(const char	*font_name)
{
	struct font_data	*fd;
	font_t	*f;
	int		dx, dy;
	char	attr[16];
	char	*temp;
	char	*p;

	f = alloc_font();
	strcpy(f->name, font_name);
	if(!strcmp(font_name, "system"))strcpy(f->name, "misc-fixed-13");
	if(!strcmp(font_name, "fixed"))strcpy(f->name, "misc-fixed-10");
	fd = f->impl_data;
	fd->auto_dealloc = 1;
	temp = fd->x_fname;
	strcpy(temp, f->name);
	p = temp;
	p += strlen(p) - 1;
	while(p>temp && *p!='-')p--;
	if(p<=temp || *p!='-')	// hopefully wont happen (unless syntax error)
	{
		f->size = 12;
		f->attr = 0;
	}
	else
	{
		*p++ = 0;	// f->name shouldn't show size and attributes
		switch(sscanf(p, "%dx%d%s", &dx, &dy, attr))
		{
			case	1:	// only size and maybe attr
				if(sscanf(p, "%d%s", &dy, attr) == 2)	// got attr
					attr[0] = attr[0];
				else attr[0] = 0;
				dx = -1;
				break;
			case	2:	// width and size, no attributes
				attr[0] = 0;
				break;
			case	3:	// kit & kaboodle :)
				break;
			default:	// something went wrong
				fprintf(stderr, "Error while parsing font string '%s': "
					"syntax is: name-[#x]#[attr]\n", font_name);
				dx = -1;
				dy = 12;
				attr[0] = 0;
				break;
		}
		f->size = dy;
		if(!strcasecmp(attr, "bi"))
			f->attr = 'z';
		else f->attr = attr[0];
	}
	fd->xfs = x11_fs_for_font(f);
	if(!fd->xfs)
	{
		free(fd);
		free(f);
		return 0;
	}
	return f;
}

XFontSet	x11_fs_for_fname(char	*fname)
{
	font_t		*f;
	XFontSet	fs;

	f = font_from_name(fname);
	fs = ((font_data*)f->impl_data)->xfs;
	free(f);
	return fs;
}

font_t	*get_system_font()
{
	return font_from_name("system");
}

int	do_select_font()
{
//	if(yz_d->buffer_mode & YZ_BUFFER_FRONT)XSetFont(display, yz_d->gc, yz_d->fnt->fid);
//	if(yz_d->buffer_mode & YZ_BUFFER_BACK)XSetFont(display, yz_d->pgc, yz_d->fnt->fid);
}

yzSelectFont1s(char	*font_name)
{
	font_t	*nf;

	nf = font_from_name(font_name);
	if(!nf)return -1;
	yzSelectFont1f(nf);
	return 0;
}

yzSelectFont1f(font_t	*nf)
{
	XFontSet	fs;
	font_data	*fd;
	font_t	*of = 0;

	fd = nf->impl_data;
	fs = fd->xfs;

	fd->refcnt++;

	if(!yz_current_window || yz_d->is_postscript)
	{
		of = no_window_font;
		no_window_font = nf;
		no_window_x11_font = fs;
	}
	else
	{
		of = yz_current_window->font;
		yz_current_window->font = nf;
		if(memcmp(&fs,&yz_d->fnt,sizeof(fs)))
		{
			yz_d->fnt = fs;
			do_select_font();
		}
	}
	if(of && ((font_data*)of->impl_data)->auto_dealloc)free_font(of);
	return 0;
}

int	non_window_string_length(char	*s, int	n)
{
	if(!display)init_xyz();
	if(!no_window_font)yzSelectFont1s("system");
	if(no_window_font)return font_string_width(no_window_x11_font, s, n);
	else return -1;
}

font_t	*yzGetFont1s(char	*font_name)
{
	font_t	*f = font_from_name(font_name);
	font_data	*fd = f->impl_data;

	fd->auto_dealloc = 0;
	return f;
}

int	yzFreeFont(font_t	*f)
{
	free_font(f);
	return 0;
}

#endif
