/*
 * XYZ Library - window_data.h
 *
 * Copyright (C) 2001 Woodley Packard
 * 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include	<X11/Xlib.h>
#include	<X11/extensions/XShm.h>
#ifdef USE_PANGO
#include	<pango/pangoxft.h>
#endif

typedef struct window_data_item
{
	char	key[128];
	void	*data;
	struct window_data_item	*next, *prev;
}	window_data_item;

typedef struct event_queue
{
	event_t	ev;
	struct event_queue	*next, *prev;
}	event_queue;

typedef struct event_filter
{
	int	(*filter)(event_t*);
	struct event_filter	*next, *prev;
}	event_filter;

typedef struct rect_info
{
	int	x1, y1, x2, y2;
}	rect_info;

typedef struct point_info
{
	int	x, y;
}	point_info;

typedef	struct window_data
{
	Window	xw;
	Pixmap	xp;
	int		buffer_mode;
	int		have_custom_cursor;
	Cursor	xcursor;
	GC		gc, pgc;
	unsigned long		col;
	int		x, y, b;
#ifdef USE_PANGO
	PangoFontDescription	*fnt;
#else
	XFontSet	fnt;
#endif
	event_filter	*evf;
	event_queue	*evq;
	window_data_item	*wd;
	int			evqd;
	char		keypressed[1024];
	char		*buffer;
	int			have_xshm;
	XShmSegmentInfo	seginfo;

	struct
	{
		int		n, a;
		rect_info	*r;
	} cs;
	struct
	{
		int		n, a;
		point_info	*p;
	} os;

	int	is_postscript, fd;

	Atom	wm_delete_window;
}	window_data;

#ifdef USE_PANGO
typedef struct font_data
{
	int			auto_dealloc;
	PangoFontDescription *fd;
	int		refcnt;
}	font_data;
#else
typedef struct font_data
{
	char		x_fname[128];
	int			auto_dealloc;
	XFontSet	xfs;
	int		refcnt;
}	font_data;
#endif

#define	XYZ_EVM	PointerMotionMask | ExposureMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask | EnterWindowMask | LeaveWindowMask | StructureNotifyMask

int	do_select_font();
int	do_select_pen();
